/*
 * Ported from: NGramLM.cpp and NGramLM.h (Original Author: Keith Vertanen)
 */

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

public class NGramLM {
	
	private static String PORTING_DEBUG_CURRENT_LINE = ""; // For Java port debugging purposes: stores the last read line
	
	protected long						m_iTotalNGrams;			// Count of total n-grams in this LM
	protected List<NGramLMNode>			m_ppUnigrams;			// Contains nodes for each unigram in the LM, stored by unique ID.
	protected int						m_iNumUnigrams;			// How many unigram roots we have
	protected int						m_iMaxSize;				// Max size of n-gram's we hold
	protected Map<String, Integer>		m_mapWords;				// Maps word strings to their numeric ID
	protected int						m_iDebug;				// User settable debug level
	protected List<String>				m_strWords;				// Allows numeric IDs to be mapped to string (helpful for debugging)
	protected String					m_strErrorLog;			// Filename of the error log file
	protected float						m_fInvLogOfBase;		// If not 0.0, then multiple all result probs by this
	protected String					m_strStartWord;			// Sentence start word, normally <s>
	protected String					m_strEndWord;			// Sentence end word, normally </s>
	protected String					m_strUnkWord;			// Unknown word, normally <unk>
	protected boolean					m_bHasUnk;				// Does the loaded LM have the unknown word?

	/**
	 * Simple form of constructor, loads from the given sorted ARPA format LM
	 */
	public NGramLM(String strFilename) throws IOException
	{
	    this(strFilename, 0, true, 0.0f, "", 0, false);
	}
	
	public NGramLM(String strFilename, int iDebug, boolean bSorted, float fResultLogBase) throws IOException
	{
		this(strFilename, iDebug, bSorted, fResultLogBase, "", 0, false);
	}

	/**
	 * Build ourselves based on an ARPA format ASCII LM.
	 * strFilename    - file to load from, if *.gz then we use zcat to decompress the file
	 * iDebug         - debug level, 1 
	 * bSorted        - (optional) if true, higher order n-grams sorted (this makes building require less memory)
	 * fResultLogBase - (optional) if not 0.0, then we change result log probs to this base (for example 10 instead of base e)
	 * strOutFilename - (optional) output sorted/truncated n-grams to this file
  	 * iTrunc         - (optional) truncate output LM to this many decimal places
	 * bOutOnly       - (optional) if true, we don't create a tree structure (for sorting/truncation only)
	 */
	public NGramLM(String strFilename, int iDebug, boolean bSorted, float fResultLogBase, String strOutFilename,  int iTrunc, boolean bOutOnly) throws IOException
	{
		List<String> vecVocab = new ArrayList<String>(); 
		LoadModel(strFilename, vecVocab, iDebug, bSorted, fResultLogBase, strOutFilename, iTrunc, bOutOnly);
	}
	
	/**
	 * Constructor in which the caller is giving us a vocab list.  This allows various LMs
	 * and other classes to share the same word to numeric ID mapping to prevent confusion.
	 */
	public NGramLM(String strFilename, List<String> vecVocab, int iDebug, boolean bSorted, float fResultLogBase, String strOutFilename, int iTrunc, boolean bOutOnly) throws IOException
	{
		LoadModel(strFilename, vecVocab, iDebug, bSorted, fResultLogBase, strOutFilename, iTrunc, bOutOnly);
	}
	
	/**
	 * This is a helper that is called by various forms of the constructor
	 */
	private void LoadModel(String strFilename, List<String> vecVocab, int iDebug,  boolean bSorted, float fResultLogBase, String strOutFilename, int iTrunc,  boolean bOutOnly) throws IOException {
		InitMemberVars();
		
		m_iDebug = iDebug;
		String strBuffer = new String();

		if (fResultLogBase > 0.0f)
			m_fInvLogOfBase = (float) (1.0 / Math.log(fResultLogBase));
		
		int				iCurrentBlock	= 0;		// 0 = header, 1 = unigrams, 2 = bigrams, etc
		boolean			bFoundData		= false;
		int				iChunks			= 0;
		float			fProb			= 0.0f;
		float			fBOW			= 0.0f;
		String			strWord			= "";
		int				iNumUnigrams	= 0;
		double			dUniSum			= 0.0;
		NGramLMBase		pLastNGram		= null;
		NGramLMBase		pCurrentNGram	= null;
		PrintStream		fpOut			= null;
		boolean			bWarned			= false;

		List<Integer>							vecWordIDs = new ArrayList<Integer>(); // C++ type: "typedef deque<NGramWordID> DEQUE_NGRAM_WORD_ID;"
		String[]								pStartChunk = null; 
//		Iterator<Map.Entry<String, Integer>> 	iterMap; // C++ type: "typedef hash_map<string, NGramWordID, HashString>::iterator MAP_STR_NGRAM_WORD_ID_ITER;"
		ArrayList<NGramLMBase>					pVecNGrams = null; // C++ type: "typedef	vector<NGramLMBase*> VECTOR_NGRAM_LM_BASE_PTR;"
		
		pVecNGrams = new ArrayList<NGramLMBase>();

		if (strOutFilename.length() > 0)
			fpOut = new PrintStream(strOutFilename);

		if (m_iDebug > 1)
			System.out.println("Before loading LM, memory usage = " + Runtime.getRuntime().freeMemory() / 1024 + "K");

		if (m_iDebug > 1)
			System.out.println("Word ID size in bytes = " + Integer.SIZE / 8 + ", max = " + NGramLMTypes.NGRAM_WORD_ID_MAX);

		// If sorted, we don't really know how big this is likely to get
		pVecNGrams.ensureCapacity(4096);

		m_mapWords = new HashMap<String, Integer>(); // Porting note: Java requires explicit instantiation

		// See if we have an existing vocab to load into the map
		if (vecVocab.size() > 0)
		{
			int i = 0;
			
			// This mapping from ID to string has to be as big as the vocab size
			m_strWords = new ArrayList<String>(vecVocab.size());
			m_strWords.addAll(Collections.nCopies(vecVocab.size(), (String)null)); // Porting note: Required to ensure Java list elements are allocated
			
			Iterator<String> iter = vecVocab.iterator();
			while (iter.hasNext()) {
				String v = iter.next();
				m_mapWords.put(v, Integer.valueOf(i));
				m_strWords.set(i, v);
				i++;			
			}
		}
		
		BufferedReader fp = null;

		// We allow optional reading of a gzipped language model
		if (strFilename.contains(".gz"))
		{
			//String strCmd = "zcat ";
			//strCmd += strFilename;
			//fp =  popen(strCmd.c_str(), "r");
			throw new UnsupportedOperationException("gzip stream unsupported by Java port");
		}
		else
			fp = new BufferedReader(new FileReader(strFilename));

		//String pResult = null;
		
		while ((strBuffer = fp.readLine()) != null) {

			PORTING_DEBUG_CURRENT_LINE = strBuffer;
			if (!bFoundData)
			{
				if (strBuffer.indexOf("\\data\\") == 0)
					bFoundData = true;
				if (fpOut != null)
					fpOut.println(strBuffer);
                        }
			else
			{
				// Look for the end of LM line
				if (strBuffer.indexOf("\\end\\") == 0)
				{
					// Add the final set of n-grams from the last block
					if (!bSorted)
					{
						if (m_iDebug > 1)
							System.out.println("Sorting " + iCurrentBlock + "-grams, mem = " + Runtime.getRuntime().freeMemory() / 1024 + "K");

						// We need to sort all n-grams of this size before we construct the balance tree
						Collections.sort(pVecNGrams, new CompNGramLM()); // sort(pVecNGrams, CompNGramLM);

						if (!bOutOnly)
							AddNGramAll(pVecNGrams);
					}
					else
					{
						// Add the last block of n-grams 
						if (!bOutOnly)
							AddNGramBlock(pVecNGrams);
					}
					if (fpOut != null)
					{
						WriteNGramBlock(pVecNGrams, fpOut, iTrunc);
						fpOut.println(strBuffer);
					}

					pVecNGrams.clear();
					pLastNGram = null;
					pCurrentNGram = null;

					break;
				}

				switch (iCurrentBlock)
				{
					case 0:
					{
						// Still in the header
						if (fpOut != null)
							fpOut.println(strBuffer);

						if ((strBuffer.indexOf("\\1-grams:") == 0))
						{
							++iCurrentBlock;
						}
						else
						{
							if (strBuffer.indexOf("ngram 1=") == 0)
							{
								// We look for the number of unigrams so we can allocate the array that
								// holds the root of each unigram.
//								char* pStart = strBuffer;							
//								pStart += 8;
								String pStart = strBuffer.substring(8);
								iNumUnigrams = Integer.parseInt(pStart);
								m_iTotalNGrams += iNumUnigrams;
								if (!bOutOnly)
								{
									m_ppUnigrams = new ArrayList<NGramLMNode>(iNumUnigrams);
									m_ppUnigrams.addAll(Collections.nCopies(iNumUnigrams, (NGramLMNode)null)); // Porting note: Required to ensure Java list elements are allocated
									for (int i = 0; i < iNumUnigrams; ++i)
										m_ppUnigrams.set(i, null);
								}

								// Only do this if we didn't create from existing vocab list
								if (vecVocab.size() == 0) {
									m_strWords = new ArrayList<String>(iNumUnigrams);
									m_strWords.addAll(Collections.nCopies(iNumUnigrams, (String)null)); // Porting note: Required to ensure Java list elements are allocated
								}

								if (m_iDebug > 0)
									System.out.println("1-grams: " + iNumUnigrams);							
								++m_iMaxSize;
							}
							else if (strBuffer.contains("ngram "))
							{
								int iSize = 0;
								int iNum  = 0;
//								sscanf(strBuffer, "ngram %d=%d", &iSize, &iNum);
								String[] tokens = strBuffer.split(" ");
								String[] ints = tokens[1].split("=");
								iSize = Integer.parseInt(ints[0]);
								iNum = Integer.parseInt(ints[1]);
								if (m_iDebug > 0)
									System.out.println(iSize + "-grams: " + iNum);										
								++m_iMaxSize;
								m_iTotalNGrams += iNum;
							}

						}
						break;
					}
					case 1:
					{
						// Unigrams
						if (strBuffer.indexOf("\\2-grams:") == 0)
						{
							if (fpOut != null)
								fpOut.println(strBuffer);
							++iCurrentBlock;
							if (m_iDebug > 0)
							{
								System.out.println("Unigram sum = " + dUniSum);
								if (m_iDebug > 1)
									System.out.println("Reading 2-grams, mem = " + Runtime.getRuntime().freeMemory() / 1024 + "K");
								else
									System.out.println("Reading 2-grams...");
							}
						}
						else
						{
							// -2.773	!exclamation-point	-2.018
							pStartChunk = SplitBuffer(strBuffer);
							iChunks = pStartChunk.length;
							fProb = 0.0f;
							fBOW = 0.0f;
							strWord = "";
							if (iChunks >= 2)
							{
								fProb = (float) (Double.parseDouble(pStartChunk[0]) * NGramLMTypes.LN_10);
								strWord = pStartChunk[1];
								if (iChunks >= 3)
									fBOW = (float) (Double.parseDouble(pStartChunk[2]) * NGramLMTypes.LN_10);

								// Sanity check for unigrams
								dUniSum += Math.exp(fProb);

								// Add this to our internal map of strings to numeric IDs
								if (m_iNumUnigrams < NGramLMTypes.NGRAM_WORD_ID_MAX)
								{
									int iTargetIndex = m_iNumUnigrams;
									if (vecVocab.size() > 0)
									{
										// We have a vocab, so map to the same numeric ID
//										iterMap = m_mapWords.find(strWord);
//										if (iterMap != m_mapWords.end())
//											iTargetIndex = iterMap->second;
//										else
//										{
//											System.err.printf("Word '%s' is not in the vocab!");
//											break;
//										}									
										Integer v = m_mapWords.get(strWord);
										if (v != null)
											iTargetIndex = v.intValue();
										else
										{
											System.err.printf("Word '%s' is not in the vocab!");
											break;
										}									
									}
									else
									{
										// Normal case, no pre-loaded vocab
										m_mapWords.put(strWord, Integer.valueOf(m_iNumUnigrams));

										// This is optional, but allows us to map numeric IDs to the string (good for debugging)
										m_strWords.set(m_iNumUnigrams, strWord);
									}

									// The unigram information is root stored in array
									if (!bOutOnly)
									{
										m_ppUnigrams.set(iTargetIndex, new NGramLMNode(iTargetIndex, fProb, fBOW, 0));
									}

									++m_iNumUnigrams;
								}
								else
									System.err.printf("Vocab too big for internal word ID max %d!", NGramLMTypes.NGRAM_WORD_ID_MAX);

								if (fpOut != null)
								{
									NGramLMBase.WriteTruncFloat(fpOut, fProb / NGramLMTypes.LN_10, iTrunc);						
									fpOut.printf("\t%s\t", strWord);
									NGramLMBase.WriteTruncFloat(fpOut, fBOW / NGramLMTypes.LN_10, iTrunc);
									fpOut.printf("\n");
								}

								// Check if this unigram is the special unknown word
								if ((!m_bHasUnk) && (strWord.compareTo(m_strUnkWord) == 0))
								{
									if (m_iDebug > 0)
										System.out.println("Using unknown word for OOVs...");
									m_bHasUnk = true;
								}

							}
						}

						break;
					}
					default:
					{
						// 2-grams or above
						if ((strBuffer.indexOf("\\") == 0) && (strBuffer.indexOf("-grams") != 0))
						{
							if (!bSorted)
							{
								if (m_iDebug > 1)
									System.out.println("Sorting " + iCurrentBlock + "-grams, mem = " + Runtime.getRuntime().freeMemory() / 1024 + "K");

								// We need to sort all n-grams of this size before we construct the balance tree
								Collections.sort(pVecNGrams, new CompNGramLM());
								if (!bOutOnly)
									AddNGramAll(pVecNGrams);
							}
							else
							{
								// Add the last block of n-grams 
								if (!bOutOnly)
									AddNGramBlock(pVecNGrams);
							}
							if (fpOut != null)
							{
								WriteNGramBlock(pVecNGrams, fpOut, iTrunc);
								fpOut.println(strBuffer);
							}

							pVecNGrams.clear();
							pLastNGram = null;
							pCurrentNGram = null;

							++iCurrentBlock;

								if (m_iDebug > 1)
									System.out.println("Reading " + iCurrentBlock + "-grams, mem = " + Runtime.getRuntime().freeMemory() / 1024 + "K");
								else if (m_iDebug > 0)
									System.out.println("Reading " + iCurrentBlock + "-grams...");
						}
						else
						{
							pStartChunk = SplitBuffer(strBuffer);
							iChunks = pStartChunk.length;
							fProb = 0.0f;
							fBOW = 0.0f;
							vecWordIDs.clear();
							boolean bGood = true;
							if (iChunks > iCurrentBlock)
							{
								fProb = (float) (Double.parseDouble(pStartChunk[0]) * NGramLMTypes.LN_10);
								if (iChunks > iCurrentBlock + 1)
								{
									// We have a backoff weight
									fBOW = (float) (Double.parseDouble(pStartChunk[iChunks - 1]) * NGramLMTypes.LN_10);
									for (int i = 1; i < iChunks - 1; ++i)
									{
//										iterMap = m_mapWords.find(pStartChunk[i]);
//										if (iterMap != m_mapWords.end())
//											vecWordIDs.push_back(iterMap->second);
//										else
//										{
//											System.err.printf("Unknown word '%s'!", pStartChunk[i]);
//											bGood = false;
//											break;
//										}
										Integer v = m_mapWords.get(pStartChunk[i]);
										if (v != null)
											vecWordIDs.add(v.intValue());
										else
										{
											System.err.printf("Unknown word '%s'!", pStartChunk[i]);
											System.err.printf("LINE: %s", strBuffer);
											bGood = false;
											break;
										}
									}	
								}
								else
								{
									// No backoff weight
									for (int i = 1; i < iChunks; ++i)
									{
//										iterMap = m_mapWords.find(pStartChunk[i]);
//										if (iterMap != m_mapWords.end())
//											vecWordIDs.push_back(iterMap->second);
//										else
//										{
//											System.err.printf("Unknown word '%s'!", pStartChunk[i]);
//											bGood = false;
//											break;
//										}
										Integer v = m_mapWords.get(pStartChunk[i]);
										if (v != null)
											vecWordIDs.add(v.intValue());
										else
										{
											System.err.printf("Unknown word '%s'!", pStartChunk[i]);
											System.err.printf("LINE: %s", strBuffer);
											bGood = false;
											break;
										}
									}	
								}
										
								// Sanity check for length of IDs
								if (iCurrentBlock != vecWordIDs.size())
								{
									System.err.printf("in block %d but n-gram size %d!", iCurrentBlock, vecWordIDs.size());
									bGood = false;
								}

								if (bGood)
								{
									// See if we are on a new set of n-grams where the first n-1 words are no longer
									// the same as the other ones in the vector we have been creating.

									// First create a new NGram object for this line
									switch (iCurrentBlock)
									{
										case 2:
											pCurrentNGram = new NGram2(fProb, fBOW, vecWordIDs);
											break;
										case 3:
											pCurrentNGram = new NGram3(fProb, fBOW, vecWordIDs);
											break;
										case 4:
											pCurrentNGram = new NGram4(fProb, fBOW, vecWordIDs);
											break;
										case 5:
											pCurrentNGram = new NGram5(fProb, fBOW, vecWordIDs);
											break;
										case 6:
											pCurrentNGram = new NGram6(fProb, fBOW, vecWordIDs);
											break;
										default:
											pCurrentNGram = new NGram7Plus(fProb, fBOW, vecWordIDs);
											break;
									};

									// Figure out if something changed besides the very last word
									boolean bChange = false;
									if (pLastNGram != null)
									{
										// Sanity check for sorted LMs
										if ((bSorted) && (!bWarned) && (pLastNGram.GetFirstID() > pCurrentNGram.GetFirstID()))
										{
											System.err.println("WARNING : this LM does not appear to be sorted!");
											System.out.println("WARNING : this LM does not appear to be sorted!");
											System.out.println("Last    : " + pLastNGram.GetWords(m_strWords));
											System.out.println("Current : " + pCurrentNGram.GetWords(m_strWords));
											bWarned = true;										
										}

										for (int i = 0; i < pLastNGram.GetSize() - 1; ++i)
										{
											if (pLastNGram.GetID(i) != pCurrentNGram.GetID(i))
											{
												bChange = true;
												break;
											}
										}	
									}

									// See if we can build the tree for this set of n-grams
									if ((bSorted) && (bChange))
									{
										if (!bOutOnly)
											AddNGramBlock(pVecNGrams);
										if (fpOut != null)
											WriteNGramBlock(pVecNGrams, fpOut, iTrunc);
										pVecNGrams.clear();
										pVecNGrams.add(pCurrentNGram);
									}
									else
									{
										// Unsorted LMs, no previous n-gram, or we matched up to the next to last ID
										pVecNGrams.add(pCurrentNGram);
									}

									pLastNGram = pCurrentNGram;
									pCurrentNGram = null;

								} // end if (bGood)

							} // end if (iChunks > iCurrentBlock)
						}
						break;
					}
				}
			}
			
			
		}
		
		if (fpOut != null)
		{
			fpOut.close();
			fpOut = null;
		}
		
		if (m_iDebug > 0)
		{
			System.out.println("Loaded LM with max size = " + m_iMaxSize);
		}

		if (fp != null)
		    fp.close();
		
		if (m_iDebug > 1)
		{
			System.out.println("After loading LM, memory usage = " + Runtime.getRuntime().freeMemory() / 1024 + "K");
			System.out.println("Size of NGramLMNode = " + NGramLMNode.NGRAMLMNODE_SIZE_ESTIMATE);
			System.out.println("Size of PackedNodeData = " + NGramLMNode.PACKEDNODEDATA_SIZE_ESTIMATE);

			long[] iNumNodes		= new long[1];
			long[] iNumBackoff		= new long[1];
			long[] iNumLeft			= new long[1];
			long[] iNumRight		= new long[1];
			long[] iNumChild		= new long[1];
			long[] iMemEstimate		= new long[1];
			long[] iNoBackoffChild	= new long[1];
			Map<Integer,Long> mapOrder = new HashMap<Integer,Long>();

			for (int i = 0; i < m_iNumUnigrams; ++i)
			{
				if (m_ppUnigrams.get(Integer.valueOf(i)) != null)
					m_ppUnigrams.get(Integer.valueOf(i)).TrieStats(iNumNodes, iNumBackoff, iNumLeft, iNumRight, iNumChild, iMemEstimate, iNoBackoffChild, mapOrder);
			}

			System.out.println("Trie stats");
			System.out.println(" num nodes        : " + iNumNodes[0]);
			System.out.println(" num backoff      : " + iNumBackoff[0] + "\t(" + (double) iNumBackoff[0] / (double) iNumNodes[0] * 100.0 + "%)");
			System.out.println(" num left         : " + iNumLeft[0]    + "\t(" + (double) iNumLeft[0]    / (double) iNumNodes[0] * 100.0 + "%)");
			System.out.println(" num right        : " + iNumRight[0]   + "\t(" + (double) iNumRight[0]   / (double) iNumNodes[0] * 100.0 + "%)");
			System.out.println(" num child        : " + iNumChild[0]   + "\t(" + (double) iNumChild[0]   / (double) iNumNodes[0] * 100.0 + "%)");
			System.out.println(" lower mem        : " + (double) iMemEstimate[0] / 1024.0 + "K");
			System.out.println(" no backoff/child : " + iNoBackoffChild[0] + "\t(" + (double) iNoBackoffChild[0]   / (double) iNumNodes[0] * 100.0 + "%)");

			int i = 0;
			
			while (mapOrder.get(Integer.valueOf(i)) != null)
			{
				System.out.println(" " + (i + 1) + "-grams          : " + mapOrder.get(Integer.valueOf(i)).longValue() + "\t(" + (double) mapOrder.get(Integer.valueOf(i)).longValue() / (double) iNumNodes[0] * 100.0 + "%)");
				++i;
			}

		}


	}
	
	public float GetLogProb(int iWord1, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		return GetLogProb(vecIDs, pMatchSize);
	}

	public float GetLogProb(int iWord1, int iWord2, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		vecIDs.add(iWord2);
		return GetLogProb(vecIDs, pMatchSize);
	}

	public float GetLogProb(int iWord1, int iWord2, int iWord3, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		vecIDs.add(iWord2);
		vecIDs.add(iWord3);
		return GetLogProb(vecIDs, pMatchSize);
	}

	public float GetLogProb(String strWord1, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord1);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));
		
		return GetLogProb(vecIDs, pMatchSize);
	}

	public float GetLogProb(String strWord1, String strWord2, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord1);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));
		
		v = m_mapWords.get(strWord2);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord2);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		return GetLogProb(vecIDs, pMatchSize);
	}

	public float GetLogProb(String strWord1, String strWord2, String strWord3, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord1);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));
		
		v = m_mapWords.get(strWord2);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord2);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		v = m_mapWords.get(strWord3);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!", strWord3);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		return GetLogProb(vecIDs, pMatchSize);
	}

	/**
	 * Convience method, do the mapping from strings to numeric IDs
	 */
	/**
	 * Porting note: Needed to rename this method because of type erasure conflict
	 * with GetLogProb(List<Integer, int[] pMatchSize).
	 */
	public float GetLogProbForWords(List<String> vecWords, int[] pMatchSize)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Iterator<String> iter = vecWords.iterator();
		while (iter.hasNext()) {
			vecIDs.add(m_mapWords.get(iter.next()));
		}
		return GetLogProb(vecIDs, pMatchSize);
	}


	/**
	 * Get the log backoff of the given bigram context
	 */
	public float GetLogBackoff(int iWord1)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(int iWord1, int iWord2)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		vecIDs.add(iWord2);
		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(int iWord1, int iWord2, int iWord3)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		vecIDs.add(iWord1);
		vecIDs.add(iWord2);
		vecIDs.add(iWord3);
		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(String strWord1)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word '%s'!");
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(String strWord1, String strWord2)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();

		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word 1 '%s'!", strWord1);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		v = m_mapWords.get(strWord2);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word 2 '%s'!", strWord2);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(String strWord1, String strWord2, String strWord3)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Integer v = m_mapWords.get(strWord1);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word 1 '%s'!", strWord1);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		v = m_mapWords.get(strWord2);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word 2 '%s'!", strWord2);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		v = m_mapWords.get(strWord3);
		if (v == null)
		{
			if (m_bHasUnk)
				vecIDs.add(m_mapWords.get(m_strUnkWord));
			else
			{
				System.err.printf("Unable to find word 3 '%s'!", strWord3);
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
			}
		}
		else
			vecIDs.add(Integer.valueOf(v.intValue()));

		return GetLogBackoff(vecIDs);
	}

	/**
	 * Porting note: Needed to change the method name to avoid type erasure
	 * conflict with GetLogBackoff(List<Integer)
	 */
	public float GetLogBackoffForWords(List<String> vecWords)
	{
		List<Integer> vecIDs = new ArrayList<Integer>();
		Iterator<String> iter = vecWords.iterator();
		int i = 0;
		while (iter.hasNext())
		{
			String str = iter.next();
			Integer v = m_mapWords.get(str);
			if (v == null)
			{
				if (m_bHasUnk)
					vecIDs.add(m_mapWords.get(m_strUnkWord));
				else
				{
					System.err.printf("Unable to find word %d '%s'!", i, str);
					return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
				}
			}
			else
				vecIDs.add(Integer.valueOf(v.intValue()));

			++i;
		}
		return GetLogBackoff(vecIDs);
	}

	public float GetLogBackoff(List<Integer> vecIDs)
	{
		if (vecIDs.size() == 0)
		{
			System.err.println("Size was 0!");	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}

		// Make sure we aren't being asked to find something bigger than the LM size
		if ((int) vecIDs.size() > m_iMaxSize)
		{
			System.err.printf("n-gram size %d bigger than LM size %d", vecIDs.size(), m_iMaxSize);	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}

		// See if somebody is asking for prob of an OOV and we aren't an unknown word model
		int iStartContext = 0;
		if (!m_bHasUnk)
		{
			if (vecIDs.get(vecIDs.size() - 1) == NGramLMTypes.NGRAM_WORD_ID_MAX)
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;

			// Check if any of our context is OOV
			for (int i = 0; i < (int) vecIDs.size() - 1; ++i)
			{
				if (vecIDs.get(Integer.valueOf(i)) == NGramLMTypes.NGRAM_WORD_ID_MAX)
					iStartContext = i + 1;
			}
		}

		int				iUnigramID		= 0;
		NGramLMNode		pNode			= null;
		int[]			pIDs			= null;
		int				iSize			= 0;

		iUnigramID = vecIDs.get(iStartContext).intValue();	

		if (iUnigramID >= m_iNumUnigrams)
		{
			System.err.printf("First id %d out of range!", iUnigramID);	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;	
		}

		pNode = m_ppUnigrams.get(iUnigramID);
		if (pNode == null)
		{
			System.err.printf("No unigram with that ID!");	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}
		
		// Copy the IDs into a simple array, this give us fast random access while we search the tree
		pIDs = null;
		iSize = (int) vecIDs.size() - iStartContext;
		pIDs = new int[iSize];
		int i = 0;
		Iterator<Integer> iter = vecIDs.iterator();
		while (iter.hasNext()) {
			pIDs[i] = iter.next().intValue();
			++i;
		}

		float[] fBackoff = new float[1];
		float[] fProb = new float[1];

		if (!pNode.Find(pIDs, iSize, 0, fBackoff, fProb, true))
		{
			pIDs = null;
			return (float) 0.0;
		}

	/* This made things worse in trigram rescoring comapred to just returning BOW of 0.0
		while ((iIndex < (int) iSize) && (!pNode->Find(pIDs, iSize, iIndex, &fBackoff, &fProb, true)))
		{	
			++iIndex;

			// Advance to looking in the leading unigram given the shorter sequence
			if (iIndex < (int) iSize)
			{
				// Reset the iterator to the new location
				iUnigramID	= pIDs[iIndex];

				if (iUnigramID >= m_iNumUnigrams)
				{
					ERR("Out of range unigram ID %d, total unigrams %d!", iUnigramID, m_iNumUnigrams);	
					return (float) NGRAM_LM_LOG_PROB_ZERO;				
				}
				pNode = m_ppUnigrams[iUnigramID];
			}

			fBackoff = 0.0f;
			fProb = 0.0f;
		}

		// Check if the search completely failed, this shouldn't happen
		if (iIndex == (int) iSize)
		{
			ERR("Search failed to match anything!");	
			return (float) NGRAM_LM_LOG_PROB_ZERO;
		}
	*/

		if (m_fInvLogOfBase > 0.0f)
			return (float) (fBackoff[0] * m_fInvLogOfBase);

		return fBackoff[0];
	}


	// Fullfill our class destiny, calculate the log prob of the n-gram given
	// by the passed in unique IDs.  
	//  pMatchSize   - (optional) the size of our matching n-gram
	//  pLogBackoff  - (optional) backoff weight of matched context
	public float GetLogProb(List<Integer> vecIDs, int[] pMatchSize)
	{
		float			fResult			= (float) 0.0;
		float[]			fProb			= new float[]{0.0f};
		float[]			fBackoff		= new float[]{0.0f};
		int				iIndex			= 0;

		if (pMatchSize != null)
			pMatchSize[0] = 0;

		if (vecIDs.size() == 0)
		{
			System.err.println("Size was 0!");	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}

		// Make sure we aren't being asked to find something bigger than the LM size
		if ((int) vecIDs.size() > m_iMaxSize)
		{
			System.err.printf("n-gram size %d bigger than LM size %d", vecIDs.size(), m_iMaxSize);	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}

		// See if somebody is asking for prob of an OOV and we aren't an unknown word model
		int iStartContext = 0;
		if (!m_bHasUnk)
		{
			if (vecIDs.get(vecIDs.size() - 1).intValue() == NGramLMTypes.NGRAM_WORD_ID_MAX)
				return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;

			// Check if any of our context is OOV
			for (int i = 0; i < (int) vecIDs.size() - 1; ++i)
			{
				if (vecIDs.get(i).intValue() == NGramLMTypes.NGRAM_WORD_ID_MAX)
					iStartContext = i + 1;
			}
		}

		int iUnigramID = 0;
		NGramLMNode pNode = null;
		int[] pIDs = null;
		int iSize = 0;

		if (pMatchSize != null)
			pMatchSize[0] = (int) vecIDs.size() - iStartContext;

		iUnigramID = vecIDs.get(iStartContext).intValue();	

		if (iUnigramID >= m_iNumUnigrams)
		{
			System.err.printf("First id %d out of range!", iUnigramID);	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;	
		}

		pNode = m_ppUnigrams.get(iUnigramID);
		if (pNode == null)
		{
			System.err.println("No unigram with that ID!");	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}
		
		// Copy the IDs into a simple array, this give us fast random access while we search the tree
		pIDs = null;
		iSize = (int) vecIDs.size() - iStartContext;
		pIDs = new int[iSize];
		int i = 0;
		Iterator<Integer> iter = vecIDs.iterator();
		while (iter.hasNext()) {
			pIDs[i] = iter.next().intValue();
			++i;			
		}


		// Look for this exact n-gram in the tree root at the starting unigram.
		// If we find it, we can just use the probability.  If we fail to find it,
		// we'll get back the backoff weight and then we'll have to search again.
		while ((iIndex < (int) iSize) && (!pNode.Find(pIDs, iSize, iIndex, fBackoff, fProb, false)))
		{	
			if (pMatchSize != null)
				pMatchSize[0]--;

			// We didn't find the whole n-gram, so we should have gotten the backoff weight
			// return inf fBackoff.
			fResult += fBackoff[0];

			++iIndex;

			// Advance to looking in the leading unigram given the shorter sequence
			if (iIndex < (int) iSize)
			{
				// Reset the iterator to the new location
				iUnigramID	= pIDs[iIndex];

				if (iUnigramID >= m_iNumUnigrams)
				{
					System.err.printf("Out of range unigram ID %d, total unigrams %d!", iUnigramID, m_iNumUnigrams);	
					return (float) 0.0;				
				}
				pNode = m_ppUnigrams.get(iUnigramID);
			} 
		
			fProb[0]		= (float) 0.0;
			fBackoff[0]		= (float) 0.0;

		} // end while

		pIDs = null;

		// Check if the search completely failed, this shouldn't happen
		if (iIndex == (int) iSize)
		{
			System.err.println("Search failed to match anything!");	
			return (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		}
		fResult += fProb[0];


		if (m_fInvLogOfBase > 0.0f)
			return (float) (fResult * m_fInvLogOfBase);

		return fResult;
	}

	
	
	private void InitMemberVars()
	{
		m_ppUnigrams		= null;
		m_iNumUnigrams		= 0;
		m_iMaxSize			= 0;
		m_iDebug			= 0;
		m_strWords			= null;
		m_iTotalNGrams		= 0;
		m_strErrorLog		= NGramLMTypes.DEFAULT_ERROR_LOG_FILENAME;
		m_fInvLogOfBase		= 0.0f;
		m_strStartWord		= NGramLMTypes.DEFAULT_START_WORD;
		m_strEndWord		= NGramLMTypes.DEFAULT_END_WORD;
		m_strUnkWord		= NGramLMTypes.DEFAULT_UNK_WORD;
		m_bHasUnk			= false;

		
	}
	
	/**
	 * Given a null terminated char* array, modify in-place such that the pStartChunk
	 * arrays gets a pointer to the start of each chunk and the whitespace between 
	 * chunks is changed to be nulls.  This is a fast way to parse whitespace seprated
	 * data in a text file (such as an ARPA LM).
	 * Returns the number of chunks found.
	 */
	/**
	 * Porting note: Changed method signature as a lack of pointer arithmetic in Java makes the above call-style impossible
	 */
//	int	NGramLM::SplitBuffer(char* pBuffer, char** ppStartChunk, int iMaxChunks)
	protected String[] SplitBuffer(String pBuffer)
	{
		if (pBuffer == null)
			return new String[0];

		return pBuffer.split("\\s+");
	}
	
	/**
	 * If we are reading an unsorted LM, we have one giant block of all 2-grams/3-grams etc.
	 * After sorting, we can break into blocks and then use the same code as the sorted case.
	 */
	protected void AddNGramAll(List<NGramLMBase> vecNGrams)
	{
		// Now that they are sorted, add each block in turn
		NGramLMBase	pLastNGram		= null;
		NGramLMBase	pCurrentNGram	= null;

		List<NGramLMBase> vecBlock = new ArrayList<NGramLMBase>();
		Iterator<NGramLMBase> iter = vecNGrams.iterator();

		while (iter.hasNext())
		{
			pCurrentNGram = iter.next();

			boolean bChange = false;
			if (pLastNGram != null)
			{
				for (int i = 0; i < pLastNGram.GetSize() - 1; ++i)
				{
					if (pLastNGram.GetID(i) != pCurrentNGram.GetID(i))
					{
						bChange = true;
						break;
					}
				}	
			}
			//if ((pLastNGram) && (pLastNGram->GetNextToLastID() != pCurrentNGram->GetNextToLastID()))
			if (bChange)
			{
				AddNGramBlock(vecBlock);
				vecBlock.clear();
			}
			vecBlock.add(pCurrentNGram);

			pLastNGram = pCurrentNGram;
		}
        
		// Make sure to add the last block
        if (vecBlock.size() > 0)
        {
          AddNGramBlock(vecBlock);
          vecBlock.clear();
        }
		
	}
	
	/**
	 * Used if we are writing a sorted/truncated LM during the load
	 */
	protected void WriteNGramBlock(List<NGramLMBase> vecNGrams, PrintStream fp, int iTrunc)
	{
		ListIterator<NGramLMBase> iter = vecNGrams.listIterator();
		NGramLMBase pCurrentNGram = null;

		// Don't bother writing backoff weights for longest span n-grams (they shouldn't have any)
		boolean bWriteBackoff = true;
//		if ((iter != vecNGrams.end()) && (*iter) && ((*iter)->GetSize() == m_iMaxSize))
//			bWriteBackoff = false;
		if (iter.hasNext() && vecNGrams.get(iter.nextIndex()) != null && vecNGrams.get(iter.nextIndex()).GetSize() == m_iMaxSize)
			bWriteBackoff = false;
		
		while (iter.hasNext())
		{
			pCurrentNGram = iter.next();
			if (pCurrentNGram != null)
				pCurrentNGram.Write(fp, iTrunc, bWriteBackoff, m_strWords);
		}

	}

	/**
	 * Add the block of n-grams to the appropriate place in the tree.
	 * This is called by the constructor during loading of the ARPA LM.
	 */
	protected void AddNGramBlock(List<NGramLMBase> vecNGrams)
	{
		if (vecNGrams.size() > 0)
		{
			NGramLMBase pFirst = vecNGrams.get(0);
			if (pFirst == null)
			{
				System.err.println("First n-gram was NULL!");
				return;
			}

			// Find the root from which all these n-grams are below
			int iSize = pFirst.GetSize();
			NGramLMNode pNode = null;
			int iID = pFirst.GetFirstID();
			
			// Start out based on the first word
			pNode = m_ppUnigrams.get(iID);

			int i = 0;
			// Keep going down tree except for last word
			while (i < (iSize - 1))
			{
				iID = pFirst.GetID(i);
				if (iID < pNode.m_data_iID)
				{
					if (pNode.m_data_pLeft == null)
					{
						System.err.println("Can't move left!");
						return;
					}
					pNode = pNode.m_data_pLeft;
				}
				else if (iID > pNode.m_data_iID)
				{
					if (pNode.m_data_pRight == null)
					{
						System.err.println("Can't move right!");
						return;
					}				
					pNode = pNode.m_data_pRight;
				}
				else
				{
					NGramLMNode pChild = null;
					pChild = pNode.m_pChild;
					if (pChild == null)
					{
						// This should only happen on second to last word
						if (i < (iSize - 2))
						{
							System.err.println("Shorter n-gram has NULL child!");
							System.err.println("Last read input line: '" + PORTING_DEBUG_CURRENT_LINE + "'");
							return;
						}
						else
						{
							pNode.AddChildren(vecNGrams, iSize - 1);
							return;
						}
					}
					else
					{
						// We are moving on to the next word in the n-gram
						pNode = pChild;
						++i;
					}
				}
			}	
			
			// We should never make it here!
			System.err.println("Failed to add block!");
		}
	}
	
	
	int GetVocabSize()
	{
		return m_iNumUnigrams;
	}

	int GetMaxSize()
	{
		return m_iMaxSize;
	}

	/*
	void NGramLM::LogError(char const* func, int line, char const* fmt, ...)
	{
	    va_list pvar;
	    
	    if (m_fpErrorLog == NULL)
	        return;

	    struct timeb timebuffer;
	    char* timeline = NULL;

	    ftime( &timebuffer );
	    timeline = ctime( & ( timebuffer.time ) );

	    // Format is:
	    // Wed Jun 22 10:22:00 2005
	    // 0123456789012345678901234
		string strTimeStamp = "";
	    if ((timeline != NULL) && (strlen(timeline) > 23))
	    {
	        for (int i = 4; i < 10; i++)
	            strTimeStamp += timeline[i];            
	        for (int i = 19; i < 24; i++)
	            strTimeStamp += timeline[i];            
	        strTimeStamp += " ";
	        for (int i = 11; i < 19; i++)
	            strTimeStamp += timeline[i];
	        strTimeStamp += ".";
	        char strMs[16];
	        sprintf(strMs, "%d", timebuffer.millitm);
	        if (strlen(strMs) == 1)
	            strTimeStamp += "00";
	        else if (strlen(strMs) == 2)
	            strTimeStamp += "0";
	        strTimeStamp += strMs;
	    }

		fprintf(m_fpErrorLog, "%s %s (%d) ", strTimeStamp.c_str(), func, line);

	    va_start(pvar, fmt);
	    vfprintf(m_fpErrorLog, fmt, pvar);
	    va_end(pvar);

		fprintf(m_fpErrorLog, "\n");
	    fflush(m_fpErrorLog);
	}
	*/

	void SetStartWord(String strStartWord)
	{
		m_strStartWord = strStartWord;
	}

	void SetEndWord(String strEndWord)
	{
		m_strEndWord = strEndWord;
	}

	void SetUnkWord(String strUnkWord)
	{
		m_strUnkWord = strUnkWord;
	}
    
	/**
     * Get the avareage per-word perplexity of a given sentence.
     * Convience method, just calls GetSentStats().
     */
    float GetSentLogProb(String strSent)
    {
        NGramStats stats = GetSentStats(strSent);
        return stats.fLogProb;
    }

    
	/**
	 * Get the avareage per-word perplexity of a given sentence.
	 * Convience method, just calls GetSentStats().
	 */
	float GetSentPPL(String strSent)
	{
	    NGramStats stats = GetSentStats(strSent);
	    return stats.fAvgPPL;
	}
	
	/**
	 *  Compute statistics about a sentence.
	 * Will add the sentence start/end symbol if not already present.
	 * All OOV words are mapped to unknown word for LMs that have this.
	 */
	NGramStats GetSentStats(String strSent)
	{
		NGramStats result = new NGramStats();
		result.fAvgPPL		= 0.0f;
		result.fLogProb		= 0.0f;
		result.iOOVs		= 0;
		result.iWords		= 0;
		result.iZeroProb	= 0;
		result.iSentences	= 0;

		List<Integer> vecIDs = new ArrayList<Integer>();
		String strWord = "";
		String strLastWord = "";

		int i = 0;
		char ch = '\0';

		int iUnkID = 0;
		
		// For LMs with <unk>, we can handle any OOV words.
		// Otherwise we map OOVs is flag value so we know that p(OOV | anything) = 0.0
		// and p(in vocab | OOV) = P(in vocab)
		if (m_bHasUnk)
			iUnkID = m_mapWords.get(m_strUnkWord).intValue();
		else
			iUnkID = NGramLMTypes.NGRAM_WORD_ID_MAX;

		int iLen = (int) strSent.length();

		while (i < iLen)
		{	
			ch = strSent.charAt(i);
			if ((ch == ' ') || (ch == '\t') || (ch == '\n') || (ch == '\r') || (i == (iLen - 1)))
			{
				// Special case for last non-whitespace character
				if ((i == (iLen - 1)) && ((ch != ' ') && (ch != '\t') && (ch != '\n') && (ch != '\r')))
					strWord += ch;

				if (strWord.length() > 0)
				{
					// Add <s> if not already there
					if ((vecIDs.size() == 0) && (strWord.indexOf(m_strStartWord) != 0))
					{
						vecIDs.add(Integer.valueOf(m_mapWords.get(m_strStartWord).intValue()));
					}

					Integer v = m_mapWords.get(strWord);
					if (v != null)
					{
						vecIDs.add(m_mapWords.get(strWord));
					}
					else
					{
						// Map to unknown word, even if LM was trained with <unk>
						vecIDs.add(iUnkID);

						// Only count truly as OOV in the stats if LM doesn't have <unk> word
						if (!m_bHasUnk)
							++result.iOOVs;
					}
				}
				strLastWord = strWord;
				strWord = "";
			}
			else
			{
				strWord += ch;
			}		
			++i;
		}

		// Add the </s> if we didn't hit it in the sentence
		if (strLastWord.indexOf(m_strEndWord) != 0)
			vecIDs.add(m_mapWords.get(m_strEndWord));

		if (vecIDs.size() <= 2)
			return result;

		++result.iSentences;

		// We should always have <s> and </s>, but we don't count in the num of words result
		result.iWords = (int) vecIDs.size() - 2;

		if (m_iDebug > 0)
		{
			// Print out a string representation of what we're actually estimating
			Iterator<Integer> iter = vecIDs.iterator();
			while (iter.hasNext()) {
				int iterint = iter.next().intValue();
				if (iterint == NGramLMTypes.NGRAM_WORD_ID_MAX)
					System.out.print(m_strUnkWord + " ");
				else
					System.out.print(m_strWords.get(iterint) + " ");
			}
			System.out.println();
		}

		List<Integer> vecCurrent = new ArrayList<Integer>();
		float fLogProb = 0.0f;
		int[] iMatch = new int[]{0};

		Iterator<Integer> iter = vecIDs.iterator();
		while (iter.hasNext())
		{
			vecCurrent.add(iter.next());
			if ((int) vecCurrent.size() > m_iMaxSize)
				vecCurrent.remove(0);

			// Skip startup when we just have <s>
			if (vecCurrent.size() > 1)
			{
				fLogProb = GetLogProb(vecCurrent, iMatch);

				// Don't count OOV log prob zero into the total
				if (vecCurrent.get(vecCurrent.size() - 1) != NGramLMTypes.NGRAM_WORD_ID_MAX)
				{
					// If we should get a log prob zero error, don't count in the total
					if (fLogProb < NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO_COMPARE)
						++result.iZeroProb;
					else 				
						result.fLogProb += fLogProb;
				}
				
				if (m_iDebug > 1)
				{
					if (vecCurrent.get(vecCurrent.size() - 1) == NGramLMTypes.NGRAM_WORD_ID_MAX)
						System.out.print("\tp(" + m_strUnkWord + " |");
					else
						System.out.print("\tp(" + m_strWords.get(vecCurrent.get(vecCurrent.size() - 1)) + " |");
					for (int ix = 0; ix < (int) vecCurrent.size() - 1; ++ix)
					{
						if (vecCurrent.get(ix).intValue() == NGramLMTypes.NGRAM_WORD_ID_MAX)
							System.out.print(" " + m_strUnkWord);
						else
							System.out.print(" " + m_strWords.get(vecCurrent.get(ix).intValue()));
					}
					if (iMatch[0] == 0)
						System.out.print(")\t= [OOV] ");
					else
						System.out.print(")\t= [" + iMatch[0] + "gram] ");
					if (m_fInvLogOfBase > 0.0f)
                                            System.out.printf("%.6f", Math.exp(fLogProb / m_fInvLogOfBase));
					else
						System.out.print(Math.exp(fLogProb));
					System.out.printf(" [ %.6f ]\n", fLogProb);
				}
			}
		}
		ComputeAvgPPL(result);

		if (m_iDebug > 0)
		{
			System.out.print("words = "      	+ result.iWords);
			System.out.print(", OOVs = "       	+ result.iOOVs);
			System.out.print(", zeroprobs = "  	+ result.iZeroProb);
			System.out.print(", logprob = "    	+ result.fLogProb);
			System.out.print(", ppl = "        	+ result.fAvgPPL);
			System.out.println();
		}

		return result;
	}

	/**
	 * Calculate the stats on a file full of sentences.
	 */
	NGramStats GetFileStats(String strFilename) throws IOException
	{
		NGramStats result = new NGramStats();
		result.fAvgPPL		= 0.0f;
		result.fLogProb		= 0.0f;
		result.iOOVs		= 0;
		result.iWords		= 0;
		result.iZeroProb	= 0;
		result.iSentences	= 0;
		
	    long then = System.currentTimeMillis();

		NGramStats current;
		String pBuffer;

		BufferedReader br = new BufferedReader(new FileReader(strFilename));
		while ((pBuffer = br.readLine()) != null)
		{
			if (pBuffer.length() > 0)
			{
				current = GetSentStats(pBuffer);

				result.fLogProb		+= current.fLogProb;
				result.iOOVs		+= current.iOOVs;
				result.iSentences	+= current.iSentences;
				result.iWords		+= current.iWords;
				result.iZeroProb	+= current.iZeroProb;
			}
		}
		br.close();

		ComputeAvgPPL(result);

		if (m_iDebug > 0)
		{
			System.out.print("FILE: " 			+ strFilename + ": ");
			System.out.print("sentences = "		+ result.iSentences);
			System.out.print(", words = "      	+ result.iWords);
			System.out.print(", OOVs = "       	+ result.iOOVs);
			System.out.print(", zeroprobs = "  	+ result.iZeroProb);
			System.out.print(", logprob = "    	+ result.fLogProb);
			System.out.print(", ppl = "        	+ result.fAvgPPL);
			System.out.println();
		}

		long now = System.currentTimeMillis();
		double dElapsed = (now - then) / 1000.0d;

		if (m_iDebug > 0)
			System.out.println("Hits/sec = " + (double) result.iWords / dElapsed);

		return result;
	}

	/**
	 * Helper that computes the per word perplexity number, adjusting log base if needed.
	 */
	void ComputeAvgPPL(NGramStats pStats)
	{
		if (pStats == null)
			return;
		if (m_fInvLogOfBase > 0.0f)
			pStats.fAvgPPL = (float) Math.exp(-1.0 * (double) pStats.fLogProb / ((double) pStats.iWords - (double) pStats.iOOVs - (double) pStats.iZeroProb + (double) pStats.iSentences) / m_fInvLogOfBase);
		else
			pStats.fAvgPPL = (float) Math.exp(-1.0 * (double) pStats.fLogProb / ((double) pStats.iWords - (double) pStats.iOOVs - (double) pStats.iZeroProb + (double) pStats.iSentences));
	}

	/**
	 * Get the vocab of this LM.
	 * Returns the number of items we put in the result vector.
	 */
	int GetVocab(List<VocabWordID> pVecResult)
	{
		if (pVecResult == null)
			return 0;

		pVecResult.clear();
		for (int i = 0; i < m_iNumUnigrams; ++i)
		{
			VocabWordID item = new VocabWordID();
			item.iID = i;
			item.strWord = m_strWords.get(Integer.valueOf(i));

			pVecResult.add(item);
		}

		return pVecResult.size();
	}

	int GetStartWordID()
	{
		return m_mapWords.get(m_strStartWord).intValue();
	}

	int GetEndWordID()
	{
		return m_mapWords.get(m_strEndWord).intValue();
	}

	int	GetUnkWordID()
	{
		return m_mapWords.get(m_strUnkWord).intValue();
	}

	int GetWordID(String strWord)
	{
		return m_mapWords.get(strWord).intValue();
	}

	boolean IsInVocab(String strWord)
	{
		if (m_mapWords.get(strWord) != null)
			return true;
		return false;
	}

	/**
	 * Convert a set of word IDs to a string
	 */
	String GetWords(List<Integer> pIDs, int iSize)
	{
		String strResult = "";
		if ((pIDs == null) || (iSize < 1))
			return strResult;

		for (int i = 0; i < iSize; ++i)
		{
			if (pIDs.get(i) >= m_iNumUnigrams)
			{
				System.err.printf("ID at position %d is not valid!", i);
				return "";
			}

			if (i > 0)
				strResult += " ";
			strResult += m_strWords.get(pIDs.get(i));
		}
		return strResult;
	}

	/**
	 * Convert a word ID into a string
	 */
	String GetWord(int iID)
	{
		String strResult = "";
		if (iID >= m_iNumUnigrams)
		{
			System.err.printf("ID of %d is not valid!", iID);
			return "";
		}

		strResult = m_strWords.get(iID);
		return strResult;
	}
	
}
