/*
 * Ported from: NGramLMNode.cpp and NGramLMNode.h (Original Author: Keith Vertanen)
 */

import java.util.List;
import java.util.Map;

public class NGramLMNode {
	
	//PackedNodeData			m_data;				// Packed data for this node, let us save memory in clever ways
	public byte m_data_iID;					// ID of the word
	public byte m_data_iOrder;				// Order of the n-gram, 0 = unigram, 1 = bigram, etc
	public float m_data_fLogProb;			// Log base e of the n-gram probability
	public NGramLMNode m_data_pLeft;		// Nodes at same level with a unique ID < us
	public NGramLMNode m_data_pRight;		// Nodes at same level with a unique ID > us
	
	public static final int PACKEDNODEDATA_SIZE_ESTIMATE = 15; // Porting note: Java estimation for five private fields a 4 bytes each

	
	// This two fields are only needed if we aren't the longer n-grams (~15% in testing)
	NGramLMNode				m_pChild;			// Goes to a n-gram of one word longer span
	float					m_fLogBackoff;      // Log base e of the backoff weight

	public static final int NGRAMLMNODE_SIZE_ESTIMATE = 12; // Porting note: Java estimation for three private fields a 4 bytes each
	
	/**
	 * Construct based on a list of NGramBase pointers which we will assume
	 * are sorted by unique ID.  We attempt to make a balanced tree below ourselves.
	 * We only use the IDs in the vector from [iStart, iEnd] inclusive.
	 */
	public NGramLMNode(List<NGramLMBase> vecNGrams, int iStart, int iEnd, int iOrder)
	{
		InitMemberVars();
		if (vecNGrams.size() == 0)
		{
			System.err.println("Vector of IDs was empty!");
			return;
		}
		if ((iStart < 0) || (iEnd < 0) || (iEnd < iStart) || (iEnd >= (int) vecNGrams.size()))
		{
			System.err.printf("Invalid start = %d, end = %d", iStart, iEnd);
			return;
		}

		// We track the order, not strictly needed by allows the word ID + order to serve as a unique ID to this object
		m_data_iOrder = (byte)iOrder;

		// We will become the mid point of the ID vector
		int				iMid		= ((iEnd - iStart) / 2) + iStart;
		NGramLMBase	pNGramMid	= vecNGrams.get(iMid);

		if (pNGramMid == null)
		{
			System.err.println("pNGramMid was NULL!");
			return;
		}
		int iSize		= pNGramMid.GetSize();
		if (iSize <= 0)
		{
			System.err.printf("iSize was %d!", iSize);
			return;
		}

//		m_data_iID		= pNGramMid.GetID(iSize - 1);
		m_data_iID		= (byte)pNGramMid.GetID(iSize - 1);
		m_data_fLogProb	= pNGramMid.m_fLogProb;
		m_fLogBackoff	= pNGramMid.m_fLogBackoff;

		// Create our children if there is anything in their bit of the vector
		if ((iMid - 1) >= iStart)
			m_data_pLeft = new NGramLMNode(vecNGrams, iStart, iMid - 1, iOrder);

		if ((iMid + 1) <= iEnd)
			m_data_pRight = new NGramLMNode(vecNGrams, iMid + 1, iEnd, iOrder);
	}

	public NGramLMNode(int iID, float fLogProb, float fLogBackoff, int iOrder)
	{
		InitMemberVars();

		m_data_fLogProb	= fLogProb;
//		m_data_iID		= iID;
		m_data_iID		= (byte)iID;
		m_data_iOrder	= (byte)iOrder;

		m_fLogBackoff	= fLogBackoff;
	}
	
	public NGramLMNode()
	{
		InitMemberVars();
	}

	/**
     * Looks for a n-gram matching the IDs in the passed in vector starting at iStart through the end.
     *  If we are exactly this n-gram, then iStart will be the last index in the vector and we'll
     * match this ID.  In this case, we return true and set pProb to be our probability.
     * 
     * bReturnBackoff - if true, we return the backoff of the matching node (used by trigram rescoring algorithm).
     * 
     */
	/**
	 * PORTING NOTE: The C++ function passes two float return values via the pointers pBackoff and pProb.
	 * In the Java port these pointers are changed to two float arrays. This ported Java method passes
	 * the float values as the first element in the respective float array. (POK 21 Aug 2012).
	 */
	public boolean Find(int[] pIDs, int iSize, int iStart, float[] pBackoff, float[] pProb, boolean bReturnBackoff)
	{
		if (pIDs == null)
		{
			System.err.println("pIDs was NULL!");
			return false;
		}
		if ((pBackoff == null) || (pProb == null))
		{
			System.err.println("pBackoff or pProb was NULL!");
			return false;
		}
		if ((iStart < 0) || (iStart >= iSize))
		{
			System.err.printf("iStart = %d out of range!", iStart);
			return false;
		}

		// Find out what we are looking for
		int iTargetID = pIDs[iStart];

		if (iTargetID == m_data_iID)
		{
			// We have matched at this node location

			if (iStart == (iSize - 1))
			{
				// We are at the end of the ID vector, so we are what they are looking for!
				pProb[0]		= m_data_fLogProb;			
				pBackoff[0]	= 0.0f;
				// Sometime people are looking just for the BOW
				if (bReturnBackoff)
				{
					pBackoff[0] = m_fLogBackoff;
				}

				return true;
			}
			else
			{
				// We are searching for some n-gram say ABC -> P(C | A, B).  
				// If we don't find this 3-gram in the tree, we want to pass up 
				// the backoff weight of BO(A, B) and not anything else.  It is 
				// possible that there were occurences of the 2-gram, in which 
				// case there is no BO weight.
				
				// Fill in local variable either from our member vars or the hash map
				float fBackoff = 0.0f;
				NGramLMNode pChild = null;
					fBackoff = m_fLogBackoff;
					pChild   = m_pChild;

				if (iStart == (iSize - 2))
				{
					// We have only one more word to add, so this backoff weight
					// is the one we'll need if we don't find the next one.
					pBackoff[0] = fBackoff;

				}

				if (pChild == null)
				{
					// No children, this one isn't in here.
					pProb[0] = (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
					return false;
				}
				else
				{
					return pChild.Find(pIDs, iSize, iStart + 1, pBackoff, pProb, bReturnBackoff);
				}
			}

		} // end if target is this node's ID

		else if (iTargetID < m_data_iID)
		{
			// See if we can find the target to the left at this n-gram size	
			if (m_data_pLeft == null)
			{
				// Nope, we're stuffed this n-gram isn't in here.
				pProb[0] = (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
				return false;
			}
			else
			{
				// Search to the left
				return m_data_pLeft.Find(pIDs, iSize, iStart, pBackoff, pProb, bReturnBackoff);
			}

		} // end target is to the left
		else if (iTargetID > m_data_iID)
		{
			// See if we can find the target to the right at this n-gram size
			if (m_data_pRight == null)
			{
				// Nope, we're stuffed this n-gram isn't in here.
				pProb[0] = (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
				return false;
			}
			else
			{
				// Search to the right
				return m_data_pRight.Find(pIDs, iSize, iStart, pBackoff, pProb, bReturnBackoff);
			}

		} // end target is to the right


		// Probably shouldn't get here
		pBackoff[0]	= 0.0f;
		pProb[0]		= (float) NGramLMTypes.NGRAM_LM_LOG_PROB_ZERO;
		return false;
	}

	/**
	 * Debugging output about this node
	 */
	public void Dump(List<String> pWords)
	{
		System.out.print("NODE " + pWords.get(m_data_iID));
		System.out.print(", prob = " + m_data_fLogProb / (float) NGramLMTypes.LN_10);

		// Fill in local variable either from our member vars or the hash map
		float fBackoff = 0.0f;
		NGramLMNode pChild = null;
		fBackoff = m_fLogBackoff;
		pChild   = m_pChild;

		System.out.print(", backoff = " + fBackoff / (float) NGramLMTypes.LN_10);
		
		if (m_data_pLeft == null)
			System.out.print(", left = NULL");
		else
			System.out.print(", left = " + pWords.get(m_data_pLeft.m_data_iID));

		if (m_data_pRight == null)
			System.out.print(", right = NULL");
		else
			System.out.print(", right = " + pWords.get(m_data_pRight.m_data_iID));

		if (pChild == null)
			System.out.print(", child = NULL");
		else
			System.out.print(", child = " + pWords.get(pChild.m_data_iID));
		System.out.print("\n");
	}
	
	private void InitMemberVars()
	{
//		m_data = new PackedNodeData(); // Porting note: explicit instantiation required for Java
		m_data_pLeft	= null;
		m_data_pRight	= null;
		m_data_fLogProb	= 0.0f;
		m_data_iOrder	= 0;
		m_data_iID		= 0;
		m_fLogBackoff	= 0.0f;
		m_pChild		= null;
		
	}
	
	/**
	 * Somebody wants a set of n-grams added below us (we should match everything but the last word).
	 */
	public void AddChildren(List<NGramLMBase> vecNGrams, int iOrder)
	{	
		m_pChild = new NGramLMNode(vecNGrams, 0, (int) vecNGrams.size() - 1, iOrder);
	}
	
	/**
	 * Debug method, sum up the number of nodes, and nodes with certain optional fields set.
	 */
	protected void TrieStats(long[] pNumNodes, long[] pNumBackoff, long[] pNumLeft, long[] pNumRight, long[] pNumChild, long[] pMemEstimate, long[] pNoBackoffChild, Map<Integer,Long> pMapOrder)
	{
		float fBackoff = 0.0f;
		NGramLMNode pChild = null;

		// Lower bound on memory requires
		pMemEstimate[0] += Integer.SIZE / 8;
		pMemEstimate[0] += Float.SIZE / 8;

		pNumNodes[0]++;

//		(*pMapOrder)[m_data.iOrder]++;
		Long l = pMapOrder.get((int)m_data_iOrder);
		// Porting note: C++ STL hash_map [] operator creates the default object if it does not exist
		// We simulate this behaviour in Java below
		if (l == null) {
			l = Long.valueOf(0L);
		}
		pMapOrder.put((int)m_data_iOrder, Long.valueOf(l.longValue() + 1));

		fBackoff = m_fLogBackoff;
		pChild	 = m_pChild;

		if (Math.abs((float) fBackoff) > NGramLMTypes.NGRAM_LM_ZERO_EPSILON)
		{
			pNumBackoff[0]++;
			pMemEstimate[0] += Float.SIZE / 8;
		}
		else
		{
			if (pChild == null)
				pNoBackoffChild[0]++;
		}

		if (m_data_pLeft != null)
		{
			pNumLeft[0]++;
			m_data_pLeft.TrieStats(pNumNodes, pNumBackoff, pNumLeft, pNumRight, pNumChild, pMemEstimate, pNoBackoffChild, pMapOrder);
			pMemEstimate[0] += NGRAMLMNODE_SIZE_ESTIMATE;
		}
		if (m_data_pRight != null)
		{
			pNumRight[0]++;
			m_data_pRight.TrieStats(pNumNodes, pNumBackoff, pNumLeft, pNumRight, pNumChild, pMemEstimate, pNoBackoffChild, pMapOrder);
			pMemEstimate[0] += NGRAMLMNODE_SIZE_ESTIMATE;
		}
		if (pChild != null)
		{
			pNumChild[0]++;
			pChild.TrieStats(pNumNodes, pNumBackoff, pNumLeft, pNumRight, pNumChild, pMemEstimate, pNoBackoffChild, pMapOrder);
			pMemEstimate[0] += NGRAMLMNODE_SIZE_ESTIMATE;
		}

	}

	
}
